// ============================================================================
// Project:	Deductive Database
// Filename:	bench_j1axy_5.cpp
// Purpose:	Bottom-Up Evaluation for Join1, a(X,Y), E: no goto, with dupEl
// Last Change:	18.05.2016
// Language:	C++
// EMail:	brass@informatik.uni-halle.de
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Feldschloesschen 15, D-06120 Halle (Saale), GERMANY
// Copyright:	(c) 2015-2016 by Stefan Brass
// License:	See file "LICENSE" for copying conditions.
// Note:	There is no warranty at all - this code may contain bugs.
// ============================================================================


//=============================================================================
// Interface:
//=============================================================================

#include "ver.h"
#include "str.h"
#include "err.h"
#include "check.h"
#include "idiv.h"
#include "perf.h"
#include "stack.h"
#include "list_2.h"
#include "cur_2.h"
#include "rel_n_n.h"
#include "cur_n_n.h"
#include "pred.h"
#include "load.h"
#include "out.h"

#include "bench_j1axy_5.h"


//=============================================================================
// Constructor, Destructor:
//=============================================================================

//-----------------------------------------------------------------------------
// Constructor (initializes subobjects):
//-----------------------------------------------------------------------------

bench_j1axy_5_c::bench_j1axy_5_c() :

		// Data structures for relations:
		c2_bf("c2_bf"),
		c3_ff("c3_ff"),
		c4_bf("c4_bf"),
		d1_ff("d1_ff"),
		d2_bf("d2_bf"),

		// Relation for storing derived b1 tuples (two IDB lit):
		b1_fb("b1_fb"),
		// Note that columns 1 and 2 are exchanged,
		// because c1 is always input to the mapping, and c2 output.

		// Sets for duplicate checks:
		b1_bb("b1_bb"),
		b2_bb("b2_bb"),
		c1_bb("c1_bb"),
		a_bb("a_bb")
{
	// Set Magic number:
	CHECK_CODE(Magic = BENCH_J1AXY_5_MAGIC);
}

//-----------------------------------------------------------------------------
// Destructor:
//-----------------------------------------------------------------------------

bench_j1axy_5_c::~bench_j1axy_5_c()
{
	// Check this object:
	CHECK_VALID("bench_j1axy_5_c::destructor");

	// Make this object invalid:
	CHECK_CODE(Magic = 0);
}


//=============================================================================
// Public Methods:
//=============================================================================

//-----------------------------------------------------------------------------
// load: Load data file with EDB facts.
//-----------------------------------------------------------------------------


bool bench_j1axy_5_c::load(str_t filename) {

	// Check parameter:
	CHECK(filename != STR_NULL, "bench_j1axy_5_c::load: filename is null");

	// Check object:
	CHECK_VALID("bench_j1axy_5_c::load");

	// Loader object (will contain data about predicates):
	load_c loader;

	// Predicates (input to loader):
	pred_c c2_p("c2", ARGTYPE_INT, ARGTYPE_INT);
	pred_c c3_p("c3", ARGTYPE_INT, ARGTYPE_INT);
	pred_c c4_p("c4", ARGTYPE_INT, ARGTYPE_INT);
	pred_c d1_p("d1", ARGTYPE_INT, ARGTYPE_INT);
	pred_c d2_p("d2", ARGTYPE_INT, ARGTYPE_INT);

	// Assign relations to predicates:
	c2_p.add_rel(&c2_bf);
	c3_p.add_rel(&c3_ff);
	c4_p.add_rel(&c4_bf);
	d1_p.add_rel(&d1_ff);
	d2_p.add_rel(&d2_bf);

	// Add predicate definitions to loader:
	loader.add(&c2_p);
	loader.add(&c3_p);
	loader.add(&c4_p);
	loader.add(&d1_p);
	loader.add(&d2_p);

	// Now do the load:
	return loader.load(filename);
}


//-----------------------------------------------------------------------------
// eval: Do bottom-up evaluation.
//-----------------------------------------------------------------------------

long bench_j1axy_5_c::eval()
{
	// Input program:
	//
	// 1: answer(A,B) :- a(A,B).
	// 2: a(A,B) :- b1(A,C), b2(C,B).
	// 3: b1(A,B) :- c1(A,C), c2(C,B).
	// 4: b2(A,B) :- c3(A,C), c4(C,B).
	// 5: c1(A,B) :- d1(A,C), d2(C,B).


	// Variables for storing data values (columns of derived tuples):
	int v5_1 = -1;
	int v5_2 = -1;
	int v3_2 = -1;
	int v4_1 = -1;
	int v4_2 = -1;
	int v2_1 = -1;

	// Cursors for looping over relations:
	cur_2_c d1_r5_cur(&d1_ff);
	cur_n_n_c d2_r5_cur(&d2_bf);
	cur_n_n_c c2_r3_cur(&c2_bf);
	cur_2_c c3_r4_cur(&c3_ff);
	cur_n_n_c c4_r4_cur(&c4_bf);
	cur_n_n_c b1_r2_cur(&b1_fb);

	// number of answers found:
	long num_results = 0;

	// --------------------------------------------------------------------
	// Jump labels:
	// label 1: applying rule 5
	//		5: c1(A,B) :- d1(A,C), d2(C,B).
	//		-> c1(v5_1,v5_2)
	// label 2: using c1(v5_1,v5_2) with literal 1 in rule 3
	//		3: b1(A,B) :- c1(A,C), c2(C,B).
	//		-> b1(v5_1,v3_2)
	// label 3: applying rule 4
	//		4: b2(A,B) :- c3(A,C), c4(C,B).
	//		-> b2(v4_1,v4_2)
	// label 4: using b2(v4_1,v4_2) with literal 2 in rule 2
	//		2: a(A,B) :- b1(A,C), b2(C,B).
	//		-> a(v2_1,v4_2)
	// label 5: using a(v2_1,v4_2) with literal 1 in rule 1
	//		1: answer(A,B) :- a(A,B).
	//		-> answer(v2_1,v4_2)
	// --------------------------------------------------------------------



	// Start Rule 5: c1(A,B) :- d1(A,C), d2(C,B).
	// Derives  c1(v5_1,v5_2)
	// label_1:
	d1_r5_cur.open();
	while (d1_r5_cur.fetch()) {
		v5_1 = d1_r5_cur.c1();
		d2_r5_cur.open(d1_r5_cur.c2());
		// label_7:
		while (d2_r5_cur.fetch()) {
			v5_2 = d2_r5_cur.c2();
			if(!c1_bb.insert(v5_1,v5_2))
				continue;
			// Push backtrack label to continuation
			//backtrack_stack.push(PUSH7);
			// goto label_2;
	
			// Using c1(v5_1,v5_2) for literal 1 in
			// rule 3:  b1(A,B) :- c1(A,C), c2(C,B).
			// Derives  b1(v5_1,v3_2)
			// label_2:
			c2_r3_cur.open(v5_2);
			// label_8:
			while (c2_r3_cur.fetch()) {
				v3_2 = c2_r3_cur.c2();
				// Push backtrack label to continuation
				// backtrack_stack.push(8);
				//
				// Derived b1(v5_1,v3_2)
				if(!b1_bb.insert(v5_1,v3_2))
					continue;
				b1_fb.insert(v3_2, v5_1);
			}
		}
	}


	// Start rule 4:   b2(A,B) :- c3(A,C), c4(C,B).
	// Derives:        b2(v4_1,v4_2)
	// label_3:
	c3_r4_cur.open();
	while (c3_r4_cur.fetch()) {
		v4_1 = c3_r4_cur.c1();
		c4_r4_cur.open(c3_r4_cur.c2());
		// Continuation for label 3
		// label_9:
		while (c4_r4_cur.fetch()) {
			v4_2 = c4_r4_cur.c2();
			if(!b2_bb.insert(v4_1,v4_2))
				continue;
			// Push backtrack label to continuation
			//backtrack_stack.push(PUSH9);
			// Labels for next rules
			//goto label_4;
			// Using b2(v4_1,v4_2) with literal 2 in rule 2
			//           a(A,B) :- b1(A,C), b2(C,B).
			// Derives:  a(v2_1,v4_2)
			// label_4:
			b1_r2_cur.open(v4_1);
			// Continuation for label 4
			//label_10:
			while (b1_r2_cur.fetch()) {
				v2_1 = b1_r2_cur.c2(); // Really column 1!
				// Push backtrack label to continuation
				//backtrack_stack.push(PUSH10);
				// Labels for next rules
				//goto label_5;
				// Using a(v2_1,v4_2) with literal 1 in rule 1
				//             answer(A,B) :- a(A,B).
				// Derives:    answer(v2_1,v4_2)
				//label_5:
				if(!a_bb.insert(v2_1,v4_2))
					continue;


				// answer(v2_1,v4_2)
				if(v2_1 < 0 || v4_2 < 0)
					return -1;
				num_results++;
				out_c::answer_2(num_results, v2_1, v4_2);
			}
		}
	}

	// Print relation statistics if requested:
	if(out_c::show_rel()) {
		out_c::rel_size_headline();
		out_c::rel_size_nonstd("c2_bf", c2_bf.num_rows(),
					c2_bf.num_pages());
		out_c::rel_size(&c3_ff);
		out_c::rel_size_nonstd("c4_bf", c4_bf.num_rows(),
					c4_bf.num_pages());
		out_c::rel_size(&d1_ff);
		out_c::rel_size_nonstd("d2_bf", d2_bf.num_rows(),
					d2_bf.num_pages());
		out_c::rel_size_nonstd("b1_fb", b1_fb.num_rows(),
					b1_fb.num_pages());
		out_c::rel_size(&b1_bb);
		out_c::rel_size(&b2_bb);
		out_c::rel_size(&c1_bb);
		out_c::rel_size(&a_bb);
		out_c::rel_size_done();
	}

	// Return number of result tuples:
	return num_results;
}

//-----------------------------------------------------------------------------
// file: List of data files.
//-----------------------------------------------------------------------------

str_t bench_j1axy_5_c::file(int i) const
{
	switch(i) {
		case 0:
			return VER_FILE_JOIN1_TEST;
		case 1:
			return VER_FILE_JOIN1_10K;
		case 2:
			return VER_FILE_JOIN1_50K;
		case 3:
			return VER_FILE_JOIN1_250K;
		default:
			return STR_NULL;
	}
}


//-----------------------------------------------------------------------------
// result: List of correct results of eval method.
//-----------------------------------------------------------------------------

long bench_j1axy_5_c::result(int i) const
{
	switch(i) {
		case 0: return 4;
		case 1: return -1;
		case 2: return -1;
		case 3: return -1;
		default: return -1;
	}
}


//=============================================================================
// Debugging Support:
//=============================================================================


//-----------------------------------------------------------------------------
// check: Check the integrity of this object, return error message or NULL.
//-----------------------------------------------------------------------------

#ifdef	VER_DEBUG
str_t	bench_j1axy_5_c::check() const
{
	// Check magic number:
	if(Magic != BENCH_J1AXY_5_MAGIC)
		return "wrong magic number";

	// Check sub-objects (relations for EDB predicates):
	CHECK_SUBOBJ(c2_bf, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(c3_ff, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(c4_bf, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(d1_ff, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(d2_bf, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(b1_fb, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(b1_bb, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(b2_bb, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(c1_bb, "bench_j1axy_5_c::check");
	CHECK_SUBOBJ(a_bb, "bench_j1axy_5_c::check");

	// Ok:
	return 0;
}

#endif


//-----------------------------------------------------------------------------
// dump: Show data structure.
//-----------------------------------------------------------------------------

#ifdef	VER_DUMP
void bench_j1axy_5_c::dump(str_t headline) const
{
	// Headline:
	if(headline == STR_NULL)
		headline = "Evaluation of Join1 a(X,Y) Example, Ver. 5";
	err_c::dump_begin(headline);

	// Basic information:
	err_c::dump_str("ObjectSize:  ");
	err_c::dump_uint(static_cast<unsigned int>(sizeof(bench_j1axy_5_c)));
	err_c::dump_str(" Bytes");
	err_c::dump_nl();
	err_c::dump_nl();

	// Show number of rows in all EDB relations:
	err_c::dump_str("Number of rows in EDB relations:  ");
	err_c::dump_nl();
	err_c::dump_str("    c2_bf: ");
	err_c::dump_int(c2_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    c3_ff: ");
	err_c::dump_int(c3_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    c4_bf: ");
	err_c::dump_int(c4_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    d1_ff: ");
	err_c::dump_int(d1_ff.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    d2_bf: ");
	err_c::dump_int(d2_bf.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    b1_fb: ");
	err_c::dump_int(b1_fb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    b1_bb: ");
	err_c::dump_int(b1_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    b2_bb: ");
	err_c::dump_int(b2_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_str("    c1_bb: ");
	err_c::dump_int(c1_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();
	err_c::dump_str("    a_bb: ");
	err_c::dump_int(a_bb.num_rows());
	err_c::dump_nl();
	err_c::dump_nl();

#ifdef EVAL_J1A_FF_D_PRINT
	// Show data for relation c3_ff:
	err_c::dump_str("Relation c3_ff:");
	err_c::dump_nl();
	cur_2_c c3_cur(&c3_ff);
	c3_cur.open();
	int i = 1;
	while(c3_cur.fetch()) {
		err_c::dump_str("  ROW ");
		err_c::dump_str(str_fill(i, 3));
		err_c::dump_int(i);
		err_c::dump_str(": (");
		int c1 = c3_cur.c1();
		err_c::dump_int(c1);
		err_c::dump_str(", ");
		int c2 = c3_cur.c2();
		err_c::dump_int(c2);
		err_c::dump_str(")");
		err_c::dump_nl();
		i++;
	}
	c3_cur.close();
	err_c::dump_nl();
#endif

	// Done:
	err_c::dump_end();
}
#endif
